/*
 * Synopsis:
 * 	double disrd(int stream, int *retval)
 *
 *	Gets a Data-is-Strings floating point number from <stream> and converts
 *	it into a double which it returns.  The number from <stream> consists of
 *	two consecutive signed integers.  The first is the coefficient, with its
 *	implied decimal point at the low-order end.  The second is the exponent
 *	as a power of 10.
 *
 *	*<retval> gets DIS_SUCCESS if everything works well.  It gets an error
 *	code otherwise.  In case of an error, the <stream> character pointer is
 *	reset, making it possible to retry with some other conversion strategy.
 *
 *	By fiat of the author, neither loss of significance nor underflow are
 *	errors.
 */

#include <pbs_config.h>   /* the master config generated by configure */

#include <assert.h>
#include <math.h>
#include <stddef.h>
#include <stdio.h>

#include "dis.h"
#include "dis_.h"
#undef disrd

double disrd(stream, retval)
    int			stream;
    int			*retval;
    {
	int		expon;
	unsigned	uexpon;
	int		locret;
	int		negate;
	unsigned	ndigs;
	unsigned	nskips;
	dis_long_double_t	ldval;

	assert(retval != NULL);
	assert(disr_commit != NULL);

	ldval = 0.0L;
	locret = disrl_(stream, &ldval, &ndigs, &nskips, DBL_DIG, 1);
	if (locret == DIS_SUCCESS) {
		locret = disrsi_(stream, &negate, &uexpon, 1);
		if (locret == DIS_SUCCESS) {
			expon = negate ? nskips - uexpon : nskips + uexpon;
			if (expon + (int)ndigs > DBL_MAX_10_EXP) {
				if (expon + (int)ndigs > DBL_MAX_10_EXP + 1) {
					ldval = ldval < 0.0L ?
					    -HUGE_VAL : HUGE_VAL;
					locret = DIS_OVERFLOW;
				} else {
				        ldval *= disp10l_(expon - 1);
					if (ldval > DBL_MAX / 10.0L) {
						ldval = ldval < 0.0L ?
						    -HUGE_VAL : HUGE_VAL;
						locret = DIS_OVERFLOW;
					} else
					        ldval *= 10.0L;
				}
			} else {
				if (expon < LDBL_MIN_10_EXP) {
					ldval *= disp10l_(expon + (int)ndigs);
					ldval /= disp10l_((int)ndigs);
				} else
				        ldval *= disp10l_(expon);
			}
		}
	}
	if ((*disr_commit)(stream, locret == DIS_SUCCESS) < 0)
	        locret = DIS_NOCOMMIT;
	*retval = locret;
	return ((double)ldval);
}
